@@ -5,6 +5,7 @@ import sys
 import ssl
 import zlib
 import os
+import brotli
 from six.moves import xrange, cStringIO
 from six.moves.urllib.parse import urlencode
 from six import print_, reraise, string_types, text_type
@@ -32,7 +33,8 @@ class ConnectionError(Exception):
             self.text = str(args[0]) if len(args) == 1 else ''
         if kwargs:
             self.text += ', ' if self.text else ''
-            self.text += ', '.join("%s=%s" % (key, val) for key, val in six.iteritems(kwargs))
+            self.kwargs_text = ', '.join('%s=%s' % (key, val) for key, val in six.iteritems(kwargs))
+            self.text += self.kwargs_text
         else:
             self.text = ''
 
@@ -42,6 +44,12 @@ class ConnectionError(Exception):
         else:
             return "URL %s" % self.url
 
+    def __repr__(self):
+        repr_str = super().__repr__()
+        if self.kwargs_text:
+            return repr_str.replace(')', ''.join([', ', self.kwargs_text, ')']))
+        return repr_str
+
 
 class RetriesExceeded(ConnectionError):
     pass
@@ -59,7 +67,8 @@ class CompatRequest(object):
     """ urllib / cookielib compatible request class.
         See also: http://docs.python.org/library/cookielib.html
     """
-    def __init__(self, url, method='GET', headers=None, payload=None):
+    def __init__(self, url, method='GET', headers=None, payload=None, params=None):
+        self.params = params
         self.set_url(url)
         self.original_host = self.url_split.host
         self.method = method
@@ -72,7 +81,7 @@ class CompatRequest(object):
             self.url_split = url
         else:
             self.url = url
-            self.url_split = URL(self.url)
+            self.url_split = URL(self.url, params=self.params)
 
     def get_full_url(self):
         return self.url
@@ -91,6 +100,10 @@ class CompatRequest(object):
         """
         return False
 
+    @property
+    def unverifiable(self):
+        return self.is_unverifiable()
+
     def get_header(self, header_name, default=None):
         return self.headers.get(header_name, default)
 
@@ -160,10 +173,12 @@ class CompatResponse(object):
     def release(self):
         return self._response.release()
 
-    def unzipped(self, gzip=True):
+    def unzipped(self, gzip=True, br=False):
         bodystr = self._response.read()
         if gzip:
             return zlib.decompress(bodystr, 16 + zlib.MAX_WBITS)
+        elif br:
+            return brotli.decompress(bodystr)
         else:
             # zlib only provides the zlib compress format, not the deflate format;
             # so on top of all there's this workaround:
@@ -195,6 +210,8 @@ class CompatResponse(object):
             ret = self.unzipped(gzip=False)
         elif content_type == 'identity':
             ret = self._response.read()
+        elif content_type == 'br':
+            ret = self.unzipped(gzip=False, br=True)
         elif content_type == 'compress':
             raise ValueError("Compression type not supported: %s" % content_type)
         else:
@@ -265,7 +282,7 @@ class UserAgent(object):
     def __del__(self):
         self.close()
 
-    def _make_request(self, url, method='GET', headers=None, payload=None):
+    def _make_request(self, url, method='GET', headers=None, payload=None, params=None):
         req_headers = self.default_headers.copy()
         if headers:
             req_headers.update(headers)
@@ -283,13 +300,13 @@ class UserAgent(object):
                 # See restkit for some example implementation
                 # TODO: Implement it
                 raise NotImplementedError
-        return CompatRequest(url, method=method, headers=req_headers, payload=payload)
+        return self.request_type(url, method=method, headers=req_headers, payload=payload, params=params)
 
     def _urlopen(self, request):
         client = self.clientpool.get_client(request.url_split)
         resp = client.request(request.method, request.url_split.request_uri,
                               body=request.payload, headers=request.headers)
-        return CompatResponse(resp, request=request, sent_request=resp._sent_request)
+        return self.response_type(resp, request=request, sent_request=resp._sent_request)
 
     def _verify_status(self, status_code, url=None):
         """ Hook for subclassing
@@ -320,7 +337,7 @@ class UserAgent(object):
         raise RetriesExceeded(url, self.max_retries, original=last_error)
 
     def urlopen(self, url, method='GET', response_codes=valid_response_codes,
-                headers=None, payload=None, to_string=False, debug_stream=None, **kwargs):
+                headers=None, payload=None, to_string=False, debug_stream=None, params=None, **kwargs):
         """ Open an URL, do retries and redirects and verify the status code
         """
         # POST or GET parameters can be passed in **kwargs
@@ -330,7 +347,7 @@ class UserAgent(object):
             elif isinstance(payload, dict):
                 payload.update(kwargs)
 
-        req = self._make_request(url, method=method, headers=headers, payload=payload)
+        req = self._make_request(url, method=method, headers=headers, payload=payload, params=params)
         for retry in xrange(self.max_retries):
             if retry > 0 and self.retry_delay:
                 # Don't wait the first time and skip if no delay specified
@@ -345,7 +362,7 @@ class UserAgent(object):
                     raise
                 except BaseException as e:
                     e.request = req
-                    e = self._handle_error(e, url=req.url)
+                    last_error = self._handle_error(e, url=req.url)
                     break # Continue with next retry
 
                 # We received a response
@@ -361,17 +378,23 @@ class UserAgent(object):
                     e.request = req
                     e.http_log = self._conversation_str(req.url, resp, payload=req.payload)
                     resp.release()
-                    e = self._handle_error(e, url=req.url)
+                    last_error = self._handle_error(e, url=req.url)
                     break # Continue with next retry
 
                 if self.cookiejar is not None:
                     self.cookiejar.extract_cookies(resp, req)
 
                 redirection = resp.headers.get('location')
+                if isinstance(redirection, six.binary_type):
+                    redirection = redirection.decode('utf-8')
                 if resp.status_code in self.redirect_resonse_codes and redirection:
                     resp.release()
-                    req.redirect(resp.status_code, redirection)
-                    continue
+                    try:
+                        req.redirect(resp.status_code, redirection)
+                        continue
+                    except Exception as e:
+                        last_error = self._handle_error(e, url=req.url)
+                        break
 
                 if not to_string:
                     return resp
@@ -381,20 +404,20 @@ class UserAgent(object):
                     try:
                         ret = resp.content
                     except Exception as e:
-                        e = self._handle_error(e, url=req.url)
+                        last_error = self._handle_error(e, url=req.url)
                         break
                     else:
                         if not ret:
                             e = EmptyResponse(url, "Empty response body received")
-                            e = self._handle_error(e, url=req.url)
+                            last_error = self._handle_error(e, url=req.url)
                             break
                         else:
                             return ret
             else:
                 e = RetriesExceeded(url, "Redirection limit reached (%s)" % self.max_redirects)
-                e = self._handle_error(e, url=url)
+                last_error = self._handle_error(e, url=url)
         else:
-            return self._handle_retries_exceeded(url, last_error=e)
+            return self._handle_retries_exceeded(url, last_error=last_error)
 
     @classmethod
     def _conversation_str(cls, url, resp, payload=None):
@@ -411,7 +434,10 @@ class UserAgent(object):
             ret = 'REQUEST: ' + url + '\n' + resp._sent_request
             if payload:
                 if isinstance(payload, six.binary_type):
-                    ret += payload.decode('utf-8') + '\n\n'
+                    try:
+                        ret += payload.decode('utf-8') + '\n\n'
+                    except UnicodeDecodeError:
+                        ret += 'UnicodeDecodeError' + '\n\n'
                 elif isinstance(payload, six.text_type):
                     ret += payload + '\n\n'
             ret += 'RESPONSE: ' + resp._response.version + ' ' + \
